import numpy as np
import pickle


def load_pickled_dynasig_df(pickle_file):
    with open(pickle_file, 'rb') as f:
        dsdf = pickle.load(f)
    return dsdf


def combine_pickled_dynasig_dfs(pickle_files_list, output_name):
    assert isinstance(pickle_files_list, list)
    assert len(pickle_files_list) >= 2
    combined_dsdf = load_pickled_dynasig_df(pickle_files_list[0])
    combined_dsdf.outname = output_name
    for pickle_file in pickle_files_list[1:]:
        dsdf = load_pickled_dynasig_df(pickle_file)
        combined_dsdf.add_other_dynasig_df(dsdf)
    combined_dsdf.save_to_file()
    return combined_dsdf


class DynaSigDF:
    """Class representing a set of Dynamical Signatures (DynaSigs), which can be outputted as a data frame.

    The DynasigDF class allows for the parallel computation of DynaSigs, stores them with minimal disk space using NumPy
    .npy binary files, and implements methods to easily combine the separate DynaSigDFs into one big dataframe for
    further analysis.

    Attributes:
        files_list (list): The list of sequence variants PDB files for which to compute the Dynamical Signatures.
        exp_measures (list): A list of experimental measures, in the same order as files_list. There can be more than
                             one measure per variant (in which case exp_measures is a list of lists).
        exp_labels (list): Labels for the experimental measures, in the order they appear in each sub_list. If there is
                           only one measure per variant, type can be str.
        n_exp (int): The number of experimental measures per variant.
        outname (str): The output name where the DynaSigDF object will be saved (with the .pickle extension).
        id_func (function): The function that gives the variant identifier from the PDB filename. By default, it gives
                            the name of the file without the path to the directory and without the extension.
        beta_values (list): The list of beta values for the Entropic Signatures computed. By default, beta=1 is the
                            only value tested.
        models (list): The list of ENM models to run. By default, only ENCoM is run.
        models_labels (list): The labels for the ENM models.
        added_atypes_list (list): Can be used to add custom atom types definitions. Needs to be of the
                                  same length as 'files_list' (some elements can be None).
        added_massdef_list (list): Same as 'added_atypes_list', for mass definition files.
        use_svib (bool): If True, the vibrational entropy will be computed and added as a potential
                         predictor variable. Defaults to False because usually the Entropic Signature is
                         enough for the model to capture the vibrational entropy.
        index_dict (dict): Dictionary of parameter combinations for every row index in the data frame.
        params_dict (dict): Dictionary of data frame row indices for every combination of parameters.
        dynasigs_masslabels (list): List of labels for the masses in the DynaSig. Every PDB file must generate the
                                    same labels.
        data_array (ndarray): Storage of all the data as a data frame (each observation on a row), using a NumPy
                              2D array.
    """

    def __init__(self, files_list, exp_measures, exp_labels, output_name, id_func=None, beta_values=None, models=None,
                 models_labels=None, added_atypes_list=None, added_massdef_list=None, use_svib=False):
        """Constructor for the DynaSigDF class.

        Args:
            files_list (list): The list of PDB files on which to compute DynaSigs.
            exp_measures (list): List of experimental measures, matching the files_list. If many experimental measures
                                 are used, has to be a list of lists: [[measure1_file1, measure2_file1],
                                 [measure1_file2, measure2_file2], ...]
            exp_labels (list): The labels for each of the experimental measures. If only one measure is supplied, can
                               be a string instead of a list.
            output_name (str): the output name to use for the saved DynaSigDF object.
            id_func (function, optional): function to be ran on the input files to generate the filename IDs (which are
                                          used to identify each DynaSig in the final dataframe). If not supplied,
                                          defaults to just the filename at the end of the path, minus the extension.
            beta_values (list, optional): If supplied, the list of beta values for the computation of DynaSigs. Positive
                                          values will use the Entropic Signature, any negative value will compute the
                                          MSF (predicted B-factors).
            models (list, optional): If supplied, list of ENM models (need to inherit from the ENM metaclass from the
                                     NRGTEN package. Defaults to ENCoM.
            models_labels (list, optional): List of model labels (to identify the model used in the dataframe).
            added_atypes_list (list, optional): Can be used to add custom atom types definitions. Needs to be of the
                                                same length as 'files_list' (some elements can be None).
            added_massdef_list (list, optional): Same as 'added_atypes_list', for mass definition files.
            use_svib (bool, optional): If True, the vibrational entropy will be computed and added as a potential
                                       predictor variable. Defaults to False because usually the Entropic Signature is
                                       enough for the model to capture the vibrational entropy.
        """
        if models is None:
            from nrgten.encom import ENCoM
            models = [ENCoM]
            models_labels = ["ENCoM"]
        if beta_values is None:
            beta_values = [1]
        if id_func is None:
            id_func = lambda x: x.split('/')[-1]
        assert isinstance(beta_values, list) and isinstance(models, list) and isinstance(models_labels, list)
        assert len(models) == len(models_labels)
        assert len(files_list) == len(exp_measures)
        if isinstance(exp_measures[0], list):
            assert isinstance(exp_labels, list)
            assert len(exp_labels) == len(exp_measures[0])
            self.n_exp = len(exp_measures[0])
        else:
            self.n_exp = 1
        if added_atypes_list is not None:
            assert len(added_atypes_list) == len(files_list) == len(added_massdef_list)
        self.files_list = files_list
        self.exp_measures = exp_measures
        self.exp_labels = exp_labels
        self.outname = output_name
        self.models = models
        self.models_labels = models_labels
        self.beta_values = beta_values
        self.id_func = id_func
        self.added_atypes_list = added_atypes_list
        self.added_massdef_list = added_massdef_list
        self.use_svib = use_svib
        self.index_dict = dict()
        self.params_dict = dict()
        self._fill_index_params_dicts()
        self.data_array = None
        self.dynasigs_masslabels = None
        self._compute_dynasig_df()
        self._pickle()

    def _pickle(self):
        self.id_func = None
        self.models = [None for x in self.models]
        with open('{}.pickle'.format(self.outname), 'wb') as f:
            pickle.dump(self, f)

    def rename(self, new_name):
        self.outname = new_name

    def save_to_file(self):
        self._pickle()

    def _fill_index_params_dicts(self):
        index = 0
        for filename_id in [self.id_func(x) for x in self.files_list]:
            self.params_dict[filename_id] = dict()
            for mod_lab in self.models_labels:
                self.params_dict[filename_id][mod_lab] = dict()
                for beta in self.beta_values:
                    self.params_dict[filename_id][mod_lab][beta] = index
                    self.index_dict[index] = [filename_id, mod_lab, beta]
                    index += 1

    def get_file_ids(self):
        return [x for x in self.params_dict]

    def get_data_array(self, file_ids, beta, mod_lab=None):
        if mod_lab is None:
            assert len(self.models_labels) == 1
            mod_lab = self.models_labels[0]
        n_add = 1  # beta value
        if self.use_svib:
            n_add += 1
        n_add += self.n_exp
        new_array = np.zeros((len(file_ids), self.data_array.shape[1]))
        for i, file_id in enumerate(file_ids):
            index = self.params_dict[file_id][mod_lab][beta]
            new_array[i] = self.data_array[index]
        return new_array

    def get_column_index(self, exp_label):
        for i, lab in enumerate(self.exp_labels):
            if lab == exp_label:
                return 1 + i

    def get_svib_index(self):
        if not self.use_svib:
            raise ValueError("Called get_svib_index() on DynaSigDF with use_svib=False")
        return 1 + self.n_exp

    def get_dynasig_indices(self):
        if self.use_svib:
            start = 2 + self.n_exp
        else:
            start = 1 + self.n_exp
        return [x for x in range(start, start + len(self.dynasigs_masslabels))]

    def _compute_dynasig_df(self):
        first_flag = True
        counter = 0
        for i, filename in enumerate(self.files_list):
            filename_id = self.id_func(filename)
            for mod, mod_lab in zip(self.models, self.models_labels):
                if self.added_atypes_list is None:
                    enm = mod(filename)
                else:
                    # This allows to have unique custom residue(s) (ligand for example) per file
                    enm = mod(filename, added_atypes=self.added_atypes_list[i],
                              added_massdef=self.added_massdef_list[i])

                # only the mass name to allow for mutations
                masslabels = [x.split('.')[-1] for x in enm.get_mass_labels()]
                if first_flag:
                    n_add = 1 # beta value
                    if self.use_svib:
                        n_add += 1
                    n_add += self.n_exp
                    self.data_array = np.zeros((np.max([x+1 for x in self.index_dict]), len(masslabels)+n_add))
                    first_flag = False
                for beta_val in self.beta_values:
                    assert counter == self.params_dict[filename_id][mod_lab][beta_val]

                    if self.dynasigs_masslabels is None:
                        self.dynasigs_masslabels = masslabels
                    else:
                        _validate_masslabels(masslabels, self.dynasigs_masslabels)

                    self.data_array[counter][0] = beta_val
                    if isinstance(self.exp_measures[i], list):
                        for j in range(self.n_exp):
                            self.data_array[counter][1+j] = self.exp_measures[i][j]
                    else:
                        self.data_array[counter][1] = self.exp_measures[i]
                    if self.use_svib:
                        self.data_array[counter][1+self.n_exp] = enm.compute_vib_entropy(beta=beta_val)
                        self.data_array[counter][2+self.n_exp:] = enm.compute_bfactors_boltzmann(beta=beta_val)
                    else:
                        self.data_array[counter][1+self.n_exp:] = enm.compute_bfactors_boltzmann(beta=beta_val)
                    counter += 1

    def add_other_dynasig_df(self, other):
        if not self._properties_are_matching(other):
            raise ValueError("Trying to combine two DynaSigDF objects with unmatched properties: {} {}".format(
                             self.outname, other.outname))
        n = np.max([x+1 for x in self.index_dict])

        for other_index in other.index_dict:
            self.index_dict[other_index + n] = other.index_dict[other_index].copy()
        for other_file_id in other.params_dict:
            if other_file_id in self.params_dict:
                raise ValueError("This file ID present in both DynaSigDFs: {}".format(other_file_id))
            self.params_dict[other_file_id] = other.params_dict[other_file_id].copy()
            for mod_lab in self.params_dict[other_file_id]:
                for beta in self.params_dict[other_file_id][mod_lab]:
                    self.params_dict[other_file_id][mod_lab][beta] += n
        self.data_array = np.concatenate((self.data_array, other.data_array))
        self.files_list = self.files_list + other.files_list
        self.exp_measures = self.exp_measures + other.exp_measures

    def _properties_are_matching(self, other):
        assert isinstance(other, DynaSigDF)
        matching = True
        if self.n_exp != other.n_exp:
            matching = False
        if self.beta_values != other.beta_values:
            matching = False
        if self.models_labels != other.models_labels:
            matching = False
        if self.use_svib != other.use_svib:
            matching = False
        return matching


def _validate_masslabels(new_labels, old_labels):
    assert len(new_labels) == len(old_labels)
    for new, old in zip(new_labels, old_labels):
        if new != old:
            raise ValueError("problem with masslabels in compute_dynasigs_df: {} not equal to {}".format(new, old))



from dynasigml.dynasig_df import load_pickled_dynasig_df
from sklearn.linear_model import Lasso, LinearRegression
from sklearn.exceptions import ConvergenceWarning
from sklearn.neural_network import MLPRegressor
import numpy as np
import random
import matplotlib.pyplot as plt
import warnings
import os
import pickle
from nrgten.encom import ENCoM


def load_dynasigml_model_from_file(filename):
    with open(filename, 'rb') as f:
        dsml = pickle.load(f)
    return dsml


class DynaSigML_Model:
    """Class used to train and test machine learning models from a saved DynaSigDF object.

        The DynaSigML_Model class can automatically train and test ML models (LASSO regression and multilayer
        perceptrons) using Dynamical Signatures saved as a DynaSigDF object. It also automatically generates a
        performance report, plots testing performance and can map the LASSO coefficients back on a PDB structure.

        Attributes:
            dynasigdf (DynaSigDF): The data frame of Dynamical Signatures and other experimental measures, as a
                                   DynaSigDF object.
            pred_cols (list): The indices of the columns (variables) to be used as predictors in the ML models.
            targ_col (int): The index of the outcome to predict.
            n (int): The total number of observations (sequence variants).
            file_ids (list): The list of file identifiers for the sequence variants (generated with id_func).
            test_ids (list): The file identifiers of sequence variants in the testing set.
            train_ids (list): The file identifiers of sequence variants in the training set.
            alphas (list): The list of different regularization strengths to be used in LASSO regression.
            n_layers (int): The number of hidden layers in multilayer perceptrons (MLPs).
            layer_size (int): The size of each MLP hidden layer.
            max_iter_lasso (int): Maximum number of LASSO iterations.
            max_iter_mlp (int): Maximum number of MLP iterations.
            lasso_stats_dict (dict): Performance of LASSO models in dictionary format.
            mlp_stats_dict (dict): Performance of MLP models in dictionary format.
            standardization (ndarray): Transformation to apply to new predictors to standardize them (mean,
                                       standard deviation for every column).
    """

    def __init__(self, dynasigdf_file, alphas=None, test_ids=None, train_ids=None, id_func=None, test_prop=0.2,
                 n_layers=2, layer_size=None, predictor_columns=None, target_column=None, max_iter_lasso=1000,
                 max_iter_mlp=200):
        """Constructor for the DynaSigML_Model class.

        Args:
            dynasigdf_file (str): The filename for the DynaSigDF object.
            alphas (list, optional): The list of different regularization strengths to be used in LASSO regression.
            test_ids (list, optional): The file identifiers of sequence variants in the testing set.
            train_ids (list, optional): The file identifiers of sequence variants in the training set. Can be omitted
                                        if test_ids is supplied, but test_ids cannot be omitted if train_ids is
                                        supplied.
            id_func: (function, optional): function to be ran on the input files to generate the filename IDs (which are
                                          used to identify each DynaSig in the final dataframe). If not supplied,
                                          defaults to just the filename at the end of the path, minus the extension.
            test_prop: (float, optional): proportion of variants to use for the testing set (random sampling), if
                                          test_ids is not supplied.
            n_layers (int, optional): The number of hidden layers in multilayer perceptrons (MLPs).
            layer_size (int, optional): The size of each MLP hidden layer.
            predictor_columns (list, optional): A list of str names of the columns to be used as predictors in the ML
                                                models. To select the whole Dynamical Signature, include 'dynasig'.
            target_column (str, optional): The str name of the column to predict.
            max_iter_lasso (int, optional): Maximum number of LASSO iterations.
            max_iter_mlp (int, optional): Maximum number of MLP iterations.
            """
        self.dynasigdf = load_pickled_dynasig_df(dynasigdf_file)
        if predictor_columns is None:
            predictor_columns = []
            for label in self.dynasigdf.exp_labels[1:]:
                predictor_columns.append(self.dynasigdf.get_column_index(label))
            for index in self.dynasigdf.get_dynasig_indices():
                predictor_columns.append(index)
        else:
            assert isinstance(predictor_columns, list)
            tmp_columns = []
            for col in predictor_columns:
                if isinstance(col, int):
                    tmp_columns.append(col)
                elif isinstance(col, str):
                    if col == 'dynasig':
                        for index in self.dynasigdf.get_dynasig_indices():
                            tmp_columns.append(index)
                    elif col == 'svib':
                        tmp_columns.append(self.dynasigdf.get_svib_index())
                    else:
                        tmp_columns.append(self.dynasigdf.get_column_index(col))
            predictor_columns = tmp_columns
        self.pred_cols = predictor_columns
        if target_column is None:
            target_column = 1
        elif isinstance(target_column, str):
            target_column = self.dynasigdf.get_column_index(target_column)
        else:
            assert isinstance(target_column, int)
        self.targ_col = target_column
        if layer_size is None:
            layer_size = len(self.pred_cols)
        if id_func is None:
            id_func = lambda x: x.split('/')[-1]
        if alphas is None:
            alphas = [2**x for x in range(-15, 1)]
        self.n = len(self.dynasigdf.files_list)
        self.file_ids = [id_func(filename) for filename in self.dynasigdf.files_list]
        if train_ids is not None:
            if test_ids is None:
                raise ValueError("Training set ids specified but not testing set ids (the reverse should be done).")
        elif test_ids is not None:
            train_ids = set([x for x in self.file_ids if x not in test_ids])
            assert len(train_ids) + len(test_ids) == self.n
        else:
            n_test = int(test_prop*self.n + 1)
            test_ids = random.sample(self.file_ids, n_test)
            train_ids = [x for x in self.file_ids if x not in test_ids]
        self.test_ids = test_ids
        self.train_ids = train_ids
        self.alphas = alphas
        self.n_layers = n_layers
        self.layer_size = layer_size
        self.max_iter_lasso = max_iter_lasso
        self.max_iter_mlp = max_iter_mlp
        self.lasso_stats_dict = dict()
        self.mlp_stats_dict = dict()
        self.standardization_dict = dict()

    def save_to_file(self, filename):
        self.id_func = None
        with open('{}.pickle'.format(filename), 'wb') as f:
            pickle.dump(self, f)

    def train_test_lasso(self):
        for beta in self.dynasigdf.beta_values:
            train_data = self.dynasigdf.get_data_array(self.train_ids, beta)
            test_data = self.dynasigdf.get_data_array(self.test_ids, beta)
            train_data, test_data, standardization = standardize_data(train_data, test_data, self.pred_cols)
            self.standardization_dict[beta] = standardization
            self.lasso_stats_dict[beta] = dict()
            for alpha in self.alphas:
                self.lasso_stats_dict[beta][alpha] = dict()
                lasso_mod = Lasso(alpha=alpha, selection='random', max_iter=self.max_iter_lasso)
                warnings.filterwarnings("error")
                try:
                    lasso_mod.fit(train_data[:, self.pred_cols], train_data[:, self.targ_col])
                    self.lasso_stats_dict[beta][alpha]['Converged'] = True
                except ConvergenceWarning:
                    warnings.filterwarnings("ignore")
                    lasso_mod.fit(train_data[:, self.pred_cols], train_data[:, self.targ_col])
                    self.lasso_stats_dict[beta][alpha]['Converged'] = False
                self.lasso_stats_dict[beta][alpha]['Training_R2'] = lasso_mod.score(train_data[:, self.pred_cols], train_data[:, self.targ_col])
                self.lasso_stats_dict[beta][alpha]['Testing_R2'] = lasso_mod.score(test_data[:, self.pred_cols], test_data[:, self.targ_col])
                self.lasso_stats_dict[beta][alpha]['model'] = lasso_mod
        warnings.filterwarnings("default")

    def train_linear_model(self):
        for beta in self.dynasigdf.beta_values:
            train_data = self.dynasigdf.get_data_array(self.train_ids, beta)
            test_data = self.dynasigdf.get_data_array(self.test_ids, beta)
            train_data, test_data, standardization = standardize_data(train_data, test_data, self.pred_cols)
            self.standardization_dict[beta] = standardization
            all_data = np.concatenate((train_data, test_data))
            mod = LinearRegression()
            mod.fit(all_data[:, self.pred_cols], all_data[:, self.targ_col])
            print(beta, adjusted_r2(mod.score(all_data[:, self.pred_cols], all_data[:, self.targ_col]), len(all_data), all_data.shape[1]-2))

    def train_test_mlp(self):
        layer_sizes = []
        for i in range(self.n_layers):
            layer_sizes.append(self.layer_size)
        layer_sizes = tuple(layer_sizes)
        for beta in self.dynasigdf.beta_values:
            train_data = self.dynasigdf.get_data_array(self.train_ids, beta)
            test_data = self.dynasigdf.get_data_array(self.test_ids, beta)
            train_data, test_data, standardization = standardize_data(train_data, test_data, self.pred_cols)
            self.standardization_dict[beta] = standardization
            self.mlp_stats_dict[beta] = dict()
            mlp_mod = MLPRegressor(layer_sizes, max_iter=self.max_iter_mlp)
            warnings.filterwarnings("error")
            try:
                mlp_mod.fit(train_data[:, self.pred_cols], train_data[:, self.targ_col])
                self.mlp_stats_dict[beta]['Converged'] = True
            except ConvergenceWarning:
                warnings.filterwarnings("ignore")
                mlp_mod.fit(train_data[:, self.pred_cols], train_data[:, self.targ_col])
                self.mlp_stats_dict[beta]['Converged'] = False
            self.mlp_stats_dict[beta]['Training_R2'] = mlp_mod.score(train_data[:, self.pred_cols], train_data[:, self.targ_col])
            self.mlp_stats_dict[beta]['Testing_R2'] = mlp_mod.score(test_data[:, self.pred_cols], test_data[:, self.targ_col])
            self.mlp_stats_dict[beta]['model'] = mlp_mod
        warnings.filterwarnings("default")

    def get_best_params_lasso(self):
        best_training = float('-Inf')
        best_testing = float('-Inf')
        best_params_training = [None, None, None]
        best_params_testing = [None, None, None]
        train_best_test = None
        best_test_model = None
        for beta in self.lasso_stats_dict:
            for alpha in self.lasso_stats_dict[beta]:
                train_r2 = self.lasso_stats_dict[beta][alpha]['Training_R2']
                test_r2 = self.lasso_stats_dict[beta][alpha]['Testing_R2']
                if train_r2 > best_training:
                    best_training = train_r2
                    best_params_training = [beta, alpha, self.lasso_stats_dict[beta][alpha]['Converged']]
                if test_r2 > best_testing:
                    best_testing = test_r2
                    best_params_testing = [beta, alpha, self.lasso_stats_dict[beta][alpha]['Converged']]
                    train_best_test = train_r2
                    best_test_model = self.lasso_stats_dict[beta][alpha]['model']
        return best_training, best_params_training, best_testing, best_params_testing, train_best_test, best_test_model

    def get_best_params_mlp(self):
        best_training = float('-Inf')
        best_testing = float('-Inf')
        best_beta_training = None
        best_beta_testing = None
        train_best_test = None
        best_test_model = None
        for beta in self.mlp_stats_dict:
            train_r2 = self.mlp_stats_dict[beta]['Training_R2']
            test_r2 = self.mlp_stats_dict[beta]['Testing_R2']
            if train_r2 > best_training:
                best_training = train_r2
                best_beta_training = beta
            if test_r2 > best_testing:
                best_testing = test_r2
                best_beta_testing = beta
                train_best_test = train_r2
                best_test_model = self.mlp_stats_dict[beta]['model']
        return best_training, best_beta_training, best_testing, best_beta_testing, train_best_test, best_test_model

    def performance_report(self):
        if self.lasso_stats_dict is not None:
            print("LASSO regression performance:")
            print("\n".join(get_performance_report_lasso(*self.get_best_params_lasso())))
        if self.mlp_stats_dict is not None:
            print("MLP regressor performance:")
            print("\n".join(get_performance_report_mlp(*self.get_best_params_mlp())))

    def get_best_beta_values(self):
        best_betas = []
        if self.lasso_stats_dict is not None:
            best_betas.append(self.get_best_beta_lasso())
        if self.mlp_stats_dict is not None:
            beta_mlp = self.get_best_beta_mlp()
            if beta_mlp != best_betas[0]:
                best_betas.append(beta_mlp)
        return best_betas

    def get_best_beta_lasso(self):
        return self.get_best_params_lasso()[3][0]

    def get_best_beta_mlp(self):
        return self.get_best_params_mlp()[3]



    def make_graphs(self, folder):
        if not os.path.isdir(folder):
            os.mkdir(folder)
        if self.lasso_stats_dict is not None:
            self._make_lasso_graphs(folder)
        if self.mlp_stats_dict is not None:
            self._make_mlp_graphs(folder)

    def _make_mlp_graphs(self, folder):
        best_training, best_beta_training, best_testing, best_beta_testing, train_best_test, best_test_model = \
        self.get_best_params_mlp()
        betas = []
        test_r2 = []
        for beta in self.mlp_stats_dict:
            betas.append(np.log(beta))
            test_r2.append(self.mlp_stats_dict[beta]['Testing_R2'])
        plt.clf()
        plt.plot(betas, test_r2)
        plt.ylim(ymin=0, ymax=1)
        plt.xlabel("Log beta")
        plt.ylabel("Predictive R²")
        plt.savefig("{}/mlp_testing_r2_beta.png".format(folder))

        test_data = self.dynasigdf.get_data_array(self.test_ids, best_beta_testing)
        train_data = self.dynasigdf.get_data_array(self.train_ids, best_beta_testing)
        train_data, test_data, standardization = standardize_data(train_data, test_data, self.pred_cols)
        self.standardization_dict[best_beta_testing] = standardization
        preds = best_test_model.predict(test_data[:, self.pred_cols])
        reals = test_data[:, self.targ_col]
        plt.clf()
        plt.scatter(preds, reals, alpha=0.5)
        plt.xlabel("Predicted value")
        plt.ylabel("Experimental value")
        text = "R² = {:.2f}\nEF10% = {:.2f}".format(best_testing, _get_ef(reals, preds))
        plt.gca().text(0.05, 0.95, text, transform=plt.gca().transAxes,
                       fontsize=14, verticalalignment='top')
        plt.savefig("{}/mlp_testing_performance.png".format(folder))
        indices = np.argsort(-preds)
        with open("{}/mlp_testing_performance.df".format(folder), "w") as f:
            f.write("variant predicted measured\n")
            for i in indices:
                f.write("{} {} {}\n".format(self.test_ids[i], preds[i], reals[i]))

    def _make_lasso_graphs(self, folder):
        best_training, best_params_training, best_testing, best_params_testing, train_best_test, best_test_model = \
        self.get_best_params_lasso()
        best_beta = best_params_testing[0]
        x = []
        test_r2 = []
        for alpha in self.lasso_stats_dict[best_beta]:
            x.append(np.log2(alpha))
            te_r2 = self.lasso_stats_dict[best_beta][alpha]['Testing_R2']
            if te_r2 < 0:
                te_r2 = 0
            test_r2.append(te_r2)
        plt.clf()
        plt.plot(x, test_r2)
        plt.ylim(ymin=0, ymax=1)
        plt.xlabel("Log₂ regularization strength")
        plt.ylabel("Predictive R²")
        plt.savefig("{}/lasso_testing_r2_alpha.png".format(folder))

        test_data = self.dynasigdf.get_data_array(self.test_ids, best_beta)
        train_data = self.dynasigdf.get_data_array(self.train_ids, best_beta)
        train_data, test_data, standardization = standardize_data(train_data, test_data, self.pred_cols)
        self.standardization_dict[best_beta] = standardization
        preds = best_test_model.predict(test_data[:, self.pred_cols])
        reals = test_data[:, self.targ_col]
        plt.clf()
        plt.scatter(preds, reals, alpha=0.5)
        plt.xlabel("Predicted value")
        plt.ylabel("Experimental value")
        text = "R² = {:.2f}\nEF10% = {:.2f}".format(best_testing, _get_ef(reals, preds))
        plt.gca().text(0.05, 0.95, text, transform=plt.gca().transAxes,
                       fontsize=14, verticalalignment='top')
        plt.savefig("{}/lasso_testing_performance.png".format(folder))
        indices = np.argsort(-preds)
        with open("{}/lasso_testing_performance.df".format(folder), "w") as f:
            f.write("variant predicted measured\n")
            for i in indices:
                f.write("{} {} {}\n".format(self.test_ids[i], preds[i], reals[i]))

        plt.clf()
        betas = []
        test_r2 = []
        for beta in self.lasso_stats_dict:
            max_r2 = 0
            for alpha in self.lasso_stats_dict[beta]:
                r2 = self.lasso_stats_dict[beta][alpha]['Testing_R2']
                if r2 > max_r2:
                    max_r2 = r2
            betas.append(np.log(beta))
            test_r2.append(max_r2)
        plt.plot(betas, test_r2)
        plt.ylim(ymin=0, ymax=1)
        plt.xlabel("Log beta")
        plt.ylabel("Best predictive R²")
        plt.savefig("{}/lasso_testing_r2_beta.png".format(folder))

    def predict_lasso(self, data_array):
        assert len(data_array[0]) == len(self.pred_cols)
        best_params = self.get_best_params_lasso()
        best_lasso_model = best_params[-1]
        best_beta = best_params[3][0]
        standardization = self.standardization_dict[best_beta]
        assert len(standardization) == len(self.pred_cols)
        for i in range(data_array.shape[1]):
            mean, sd = standardization[i]
            data_array[:, i] -= mean
            if sd > 0:
                data_array[:, i] /= sd
        return best_lasso_model.predict(data_array)

    def predict_mlp(self, data_array):
        assert len(data_array[0]) == len(self.pred_cols)
        best_params = self.get_best_params_mlp()
        best_mlp_model = best_params[-1]
        best_beta = best_params[3]
        standardization = self.standardization_dict[best_beta]
        assert len(standardization) == len(self.pred_cols)
        for i in range(data_array.shape[1]):
            mean, sd = standardization[i]
            data_array[:, i] -= mean
            if sd > 0:
                data_array[:, i] /= sd
        return best_mlp_model.predict(data_array)

    def map_coefficients(self, wt_pdb_file, output_pdb_file):
        enc = ENCoM(wt_pdb_file, solve=False)
        n = len(enc.mol.masses)
        lasso_mod = self.get_best_params_lasso()[-1]
        coefs_og = lasso_mod.coef_[-n:]
        coefs = np.array(coefs_og)
        coefs /= np.max(np.abs(coefs))
        enc.set_bfactors(np.abs(coefs))
        enc.set_occupancy(coefs)
        enc._write_to_file(output_pdb_file)
        return coefs_og





def get_performance_report_lasso(best_training, best_params_training, best_testing, best_params_testing,
                                 train_best_test, best_test_model):
    report = ["Best training R²: {:.2f} with alpha={} and beta={}".format(best_training, best_params_training[1],
                                                                              best_params_training[0])]
    report.append("Best testing (predictive) R²: {:.2f} with alpha={} and beta={} (associated training R²:{:.2f})"
                  .format(best_testing, best_params_testing[1], best_params_testing[0], train_best_test))
    return report


def get_performance_report_mlp(best_training, best_beta_training, best_testing, best_beta_testing, train_best_test,
                           best_test_model):
    report = ["Best training R²: {:.2f} with beta={}".format(best_training, best_beta_training)]
    report.append("Best testing (predictive) R²: {:.2f} with beta={} (associated training R²:{:.2f})".format(
                  best_testing, best_beta_testing, train_best_test))
    return report


def standardize_data(train_data, test_data, cols):
    standardization = np.zeros((len(cols), 2))
    count = 0
    for col in cols:
        mean = np.mean(np.concatenate((train_data[:, col], test_data[:, col])))
        sd = np.std(np.concatenate((train_data[:, col], test_data[:, col])))
        standardization[count] = [mean, sd]
        for data in [train_data, test_data]:
            data[:, col] -= mean
            if sd != 0:
                data[:, col] /= sd
        count += 1
    return train_data, test_data, standardization


def adjusted_r2(r2, n, p):
    return 1 - (1-r2)*(n-1)/(n-p-1)

def _get_ef(reals, preds, prop=0.1):
    data = np.zeros((len(reals), 2))
    data[:, 0] = reals
    data[:, 1] = preds
    data = data[data[:, 1].argsort()]
    max_index = int(len(data)*prop+1)
    thresh = sorted(data[:, 0])[-max_index+1]
    count = 0
    for i in range(max_index):
        if data[-i, 0] >= thresh:
            count += 1
    return float(count)/(max_index-1) / prop





